# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test general functionality of the BeamMe space-time
module with end-to-end integration tests."""

import numpy as np
import pytest

from beamme.core.conf import bme
from beamme.core.element_beam import generate_beam_class
from beamme.core.material import MaterialBeamBase
from beamme.core.mesh import Mesh
from beamme.mesh_creation_functions.beam_arc import create_beam_mesh_arc_segment_2d
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line
from beamme.space_time.beam_to_space_time import beam_to_space_time, mesh_to_data_arrays


def get_name(beam_class):
    """Return the identifier for the given beam object."""
    match len(beam_class.nodes_create):
        case 2:
            return "linear"
        case 3:
            return "quadratic"
        case _:
            raise TypeError("Got unexpected beam element")


@pytest.mark.parametrize("n_nodes", [2, 3])
def test_integration_space_time_straight(
    n_nodes, assert_results_close, get_corresponding_reference_file_path
):
    """Create the straight beam for the tests."""

    # Create the beam mesh in space
    beam_type = generate_beam_class(n_nodes)
    mesh = Mesh()
    create_beam_mesh_line(
        mesh, beam_type, MaterialBeamBase(), [0, 0, 0], [6, 0, 0], n_el=3
    )

    # Get the space-time mesh
    space_time_mesh, return_set = beam_to_space_time(mesh, 6.9, 5, time_start=2.5)

    # Add all sets to the mesh
    space_time_mesh.add(return_set)

    # Check the mesh data arrays
    additional_identifier = get_name(beam_type)
    mesh_data_arrays = mesh_to_data_arrays(space_time_mesh)
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier, extension="json"
        ),
        mesh_data_arrays,
    )


@pytest.mark.parametrize("n_nodes", [2, 3])
def test_integration_space_time_curved(
    n_nodes, assert_results_close, get_corresponding_reference_file_path
):
    """Create a curved beam for the tests."""

    # Create the beam mesh in space
    beam_type = generate_beam_class(n_nodes)
    mesh = Mesh()
    create_beam_mesh_arc_segment_2d(
        mesh,
        beam_type,
        MaterialBeamBase(),
        [0.5, 1, 0],
        0.75,
        0.0,
        np.pi * 2.0 / 3.0,
        n_el=3,
    )

    # Get the space-time mesh
    space_time_mesh, return_set = beam_to_space_time(mesh, 6.9, 5)

    # Add all sets to the mesh
    space_time_mesh.add(return_set)

    # Check the mesh data arrays
    additional_identifier = get_name(beam_type)
    mesh_data_arrays = mesh_to_data_arrays(space_time_mesh)
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier, extension="json"
        ),
        mesh_data_arrays,
    )


@pytest.mark.parametrize("n_nodes", [2, 3])
@pytest.mark.parametrize("couple_nodes", [False, True])
def test_integration_space_time_elbow(
    n_nodes, couple_nodes, assert_results_close, get_corresponding_reference_file_path
):
    """Create an elbow beam for the tests."""

    # Create the beam mesh in space
    beam_type = generate_beam_class(n_nodes)
    mesh = Mesh()
    mat = MaterialBeamBase()
    create_beam_mesh_line(mesh, beam_type, mat, [0, 0, 0], [1, 0, 0], n_el=3)
    create_beam_mesh_line(mesh, beam_type, mat, [1, 0, 0], [1, 1, 0], n_el=2)

    # Create the couplings
    if couple_nodes:
        mesh.couple_nodes(
            reuse_matching_nodes=True, coupling_dof_type=bme.coupling_dof.fix
        )

    # Get the space-time mesh
    space_time_mesh, return_set = beam_to_space_time(mesh, 6.9, 5, time_start=1.69)

    # Add all sets to the mesh
    space_time_mesh.add(return_set)

    # Check the mesh data arrays
    additional_identifier = get_name(beam_type) + ("_coupling" if couple_nodes else "")
    mesh_data_arrays = mesh_to_data_arrays(space_time_mesh)
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier, extension="json"
        ),
        mesh_data_arrays,
    )


@pytest.mark.parametrize("n_nodes", [2, 3])
@pytest.mark.parametrize("couple_nodes", [False, True])
@pytest.mark.parametrize("arc_length", [False, True])
def test_integration_space_time_varying_material_length(
    n_nodes,
    couple_nodes,
    arc_length,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create an elbow beam for the tests."""

    beam_type = generate_beam_class(n_nodes)

    def beam_mesh_in_space_generator(time):
        """Create the beam mesh in space generator."""
        mat = MaterialBeamBase()
        pos_y = 0.25 * (time - 1.7)

        mesh_1 = Mesh()
        create_beam_mesh_line(
            mesh_1,
            beam_type,
            mat,
            [np.sin(time), 0, 0],
            [2, pos_y, 0],
            n_el=3,
            set_nodal_arc_length=arc_length,
        )

        mesh_2 = Mesh()
        create_beam_mesh_line(
            mesh_2,
            beam_type,
            mat,
            [2, pos_y, 0],
            [2, 3, 0],
            n_el=2,
            set_nodal_arc_length=arc_length,
        )

        mesh = Mesh()
        mesh.add(mesh_1, mesh_2)
        if couple_nodes:
            mesh.couple_nodes(coupling_dof_type=bme.coupling_dof.fix)
        return mesh

    # Get the space-time mesh
    space_time_mesh, return_set = beam_to_space_time(
        beam_mesh_in_space_generator, 6.9, 5, time_start=1.69
    )

    # Add all sets to the mesh
    space_time_mesh.add(return_set)

    # Check the mesh data arrays
    additional_identifier = (
        get_name(beam_type)
        + ("_coupling" if couple_nodes else "")
        + ("_arc_length" if arc_length else "")
    )
    mesh_data_arrays = mesh_to_data_arrays(space_time_mesh)
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier, extension="json"
        ),
        mesh_data_arrays,
    )


def test_integration_space_time_named_node_set(
    assert_results_close, get_corresponding_reference_file_path
):
    """Create a straight beam and check that named node sets are handled
    correctly."""

    # Create the beam mesh in space
    mesh = Mesh()
    beam_type = generate_beam_class(2)
    create_beam_mesh_line(
        mesh, beam_type, MaterialBeamBase(), [0, 0, 0], [6, 0, 0], n_el=2
    )

    # Get the space-time mesh
    space_time_mesh, return_set = beam_to_space_time(mesh, 6.9, 3, time_start=2.5)

    # Add all sets to the mesh
    return_set["start"].name = "start"
    return_set["right"].name = "right"
    return_set["surface"].name = "surface"
    space_time_mesh.add(return_set)

    # Check the mesh data arrays
    mesh_data_arrays = mesh_to_data_arrays(space_time_mesh)
    assert_results_close(
        get_corresponding_reference_file_path(extension="json"),
        mesh_data_arrays,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Functions to create beam meshes along arcs."""

import numpy as _np

from beamme.core.conf import bme as _bme
from beamme.core.rotation import Rotation as _Rotation
from beamme.mesh_creation_functions.beam_generic import (
    create_beam_mesh_generic as _create_beam_mesh_generic,
)


def create_beam_mesh_arc_segment_via_rotation(
    mesh, beam_class, material, center, axis_rotation, radius, angle, **kwargs
):
    """Generate a circular segment of beam elements.

    The circular segment is defined via a rotation, specifying the "initial"
    triad of the beam at the beginning of the arc.

    This function exists for compatibility reasons with older BeamMe implementations.
    The user is encouraged to use the newer implementation create_beam_mesh_arc_segment_via_axis

    Args
    ----
    mesh: Mesh
        Mesh that the arc segment will be added to.
    beam_class: Beam
        Class of beam that will be used for this line.
    material: Material
        Material for this segment.
    center: _np.array, list
        Center of the arc.
    axis_rotation: Rotation
        This rotation defines the spatial orientation of the arc.
        The 3rd base vector of this rotation is the rotation axis of the arc
        segment. The segment starts in the direction of the 1st basis vector
        and the starting point is along the 2nd basis vector.
    radius: float
        The radius of the segment.
    angle: float
        The central angle of this segment in radians.

    **kwargs (for all of them look into create_beam_mesh_function)
    ----
    n_el: int
        Number of equally spaced beam elements along the line. Defaults to 1.
        Mutually exclusive with l_el.
    l_el: float
        Desired length of beam elements. Mutually exclusive with n_el.
        Be aware, that this length might not be achieved, if the elements are
        warped after they are created.

    Return
    ----
    return_set: GeometryName
        Set with the 'start' and 'end' node of the line. Also a 'line' set
        with all nodes of the line.
    """

    # Convert the input to the one for create_beam_mesh_arc_segment_via_axis
    axis = axis_rotation * [0, 0, 1]
    start_point = center + radius * (axis_rotation * [0, -1, 0])
    return create_beam_mesh_arc_segment_via_axis(
        mesh, beam_class, material, axis, center, start_point, angle, **kwargs
    )


def create_beam_mesh_arc_segment_via_axis(
    mesh,
    beam_class,
    material,
    axis,
    axis_point,
    start_point,
    angle,
    **kwargs,
):
    """Generate a circular segment of beam elements.

    The arc is defined via a rotation axis, a point on the rotation axis a starting
    point, as well as the angle of the arc segment.

    Args
    ----
    mesh: Mesh
        Mesh that the arc segment will be added to.
    beam_class: Beam
        Class of beam that will be used for this line.
    material: Material
        Material for this segment.
    axis: _np.array, list
        Rotation axis of the arc.
    axis_point: _np.array, list
        Point lying on the rotation axis. Does not have to be the center of the arc.
    start_point: _np.array, list
        Start point of the arc.
    angle: float
        The central angle of this segment in radians.

    **kwargs (for all of them look into create_beam_mesh_function)
    ----
    n_el: int
        Number of equally spaced beam elements along the line. Defaults to 1.
        Mutually exclusive with l_el.
    l_el: float
        Desired length of beam elements. Mutually exclusive with n_el.
        Be aware, that this length might not be achieved, if the elements are
        warped after they are created.

    Return
    ----
    return_set: GeometryName
        Set with the 'start' and 'end' node of the line. Also a 'line' set
        with all nodes of the line.
    """

    # The angle can not be negative with the current implementation.
    if angle <= 0.0:
        raise ValueError(
            "The angle for a beam arc segment has to be a positive number!"
        )

    # Shortest distance from the given point to the axis of rotation gives
    # the "center" of the arc
    axis = _np.asarray(axis)
    axis_point = _np.asarray(axis_point)
    start_point = _np.asarray(start_point)

    axis = axis / _np.linalg.norm(axis)
    diff = start_point - axis_point
    distance = diff - _np.dot(_np.dot(diff, axis), axis)
    radius = _np.linalg.norm(distance)
    center = start_point - distance

    # Get the rotation at the start
    # No need to check the start node here, as eventual rotation offsets in
    # tangential direction will be covered by the create beam functionality.
    tangent = _np.cross(axis, distance)
    tangent /= _np.linalg.norm(tangent)
    start_rotation = _Rotation.from_rotation_matrix(
        _np.transpose(_np.array([tangent, -distance / radius, axis]))
    )

    def get_beam_geometry(alpha, beta):
        """Return a function for the position and rotation along the beam
        axis."""

        def beam_function(xi):
            """Return a point and the triad on the beams axis for a given
            parameter coordinate xi."""
            phi = 0.5 * (xi + 1) * (beta - alpha) + alpha
            arc_rotation = _Rotation(axis, phi)
            rot = arc_rotation * start_rotation
            pos = center + arc_rotation * distance
            return (pos, rot, phi * radius)

        return beam_function

    # Create the beam in the mesh
    return _create_beam_mesh_generic(
        mesh,
        beam_class=beam_class,
        material=material,
        function_generator=get_beam_geometry,
        interval=[0.0, angle],
        interval_length=angle * radius,
        **kwargs,
    )


def create_beam_mesh_arc_segment_2d(
    mesh, beam_class, material, center, radius, phi_start, phi_end, **kwargs
):
    """Generate a circular segment of beam elements in the x-y plane.

    Args
    ----
    mesh: Mesh
        Mesh that the arc segment will be added to.
    beam_class: Beam
        Class of beam that will be used for this line.
    material: Material
        Material for this segment.
    center: _np.array, list
        Center of the arc. If the z component is not 0, an error will be
        thrown.
    radius: float
        The radius of the segment.
    phi_start, phi_end: float
        The start and end angles of the arc w.r.t the x-axis. If the start
        angle is larger than the end angle the beam faces in counter-clockwise
        direction, and if the start angle is smaller than the end angle, the
        beam faces in clockwise direction.

    **kwargs (for all of them look into create_beam_mesh_function)
    ----
    n_el: int
        Number of equally spaced beam elements along the line. Defaults to 1.
        Mutually exclusive with l_el.
    l_el: float
        Desired length of beam elements. Mutually exclusive with n_el.
        Be aware, that this length might not be achieved, if the elements are
        warped after they are created.

    Return
    ----
    return_set: GeometryName
        Set with the 'start' and 'end' node of the line. Also a 'line' set
        with all nodes of the line.
    """

    # The center point has to be on the x-y plane.
    if _np.abs(center[2]) > _bme.eps_pos:
        raise ValueError("The z-value of center has to be 0!")

    # Check if the beam is in clockwise or counter clockwise direction.
    angle = phi_end - phi_start
    axis = _np.array([0, 0, 1])
    start_point = center + radius * (_Rotation(axis, phi_start) * [1, 0, 0])

    counter_clockwise = _np.sign(angle) == 1
    if not counter_clockwise:
        # If the beam is not in counter clockwise direction, we have to flip
        # the rotation axis.
        axis = -1.0 * axis

    return create_beam_mesh_arc_segment_via_axis(
        mesh,
        beam_class,
        material,
        axis,
        center,
        start_point,
        _np.abs(angle),
        **kwargs,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test general functionality of the core mesh class
with end-to-end integration tests."""

import copy
import random
import warnings
from contextlib import nullcontext

import numpy as np
import pytest

from beamme.core.boundary_condition import BoundaryCondition
from beamme.core.conf import bme
from beamme.core.element_beam import Beam3
from beamme.core.function import Function
from beamme.core.mesh import Mesh
from beamme.core.node import NodeCosserat
from beamme.core.rotation import Rotation
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.four_c.model_importer import import_four_c_model
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line
from tests.create_test_models import create_beam_to_solid_conditions_model


def create_test_mesh(get_default_test_beam_material):
    """Create a mesh with a couple of test nodes and elements."""

    # Set the seed for the pseudo random numbers
    random.seed(0)

    # Add material to mesh.
    mesh = Mesh()
    material = get_default_test_beam_material(material_type="reissner")
    mesh.add(material)

    # Add three test nodes and add them to a beam element
    for _j in range(3):
        mesh.add(
            NodeCosserat(
                [100 * random.uniform(-1, 1) for _i in range(3)],
                Rotation(
                    [100 * random.uniform(-1, 1) for _i in range(3)],
                    100 * random.uniform(-1, 1),
                ),
            )
        )
    beam = Beam3rHerm2Line3(material=material, nodes=mesh.nodes)
    mesh.add(beam)

    # Add a beam line with three elements
    create_beam_mesh_line(
        mesh,
        Beam3rHerm2Line3,
        material,
        [100 * random.uniform(-1, 1) for _i in range(3)],
        [100 * random.uniform(-1, 1) for _i in range(3)],
        n_el=3,
    )

    return mesh


def test_integration_core_mesh_rotation(
    get_default_test_beam_material,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Check if the Mesh function rotation gives the same results as rotating
    each node it self."""

    mesh_1 = create_test_mesh(get_default_test_beam_material)
    mesh_2 = create_test_mesh(get_default_test_beam_material)

    # Set the seed for the pseudo random numbers
    random.seed(0)
    rot = Rotation(
        [100 * random.uniform(-1, 1) for _i in range(3)],
        100 * random.uniform(-1, 1),
    )
    origin = [100 * random.uniform(-1, 1) for _i in range(3)]

    for node in mesh_1.nodes:
        node.rotate(rot, origin=origin)

    mesh_2.rotate(rot, origin=origin)

    # Compare the output for the two meshes.
    assert_results_close(mesh_1, mesh_2)

    # Compare with reference results.
    assert_results_close(mesh_1, get_corresponding_reference_file_path())


def test_integration_core_mesh_rotation_individual(
    get_default_test_beam_material,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Check if the Mesh function rotation gives the same results as rotating
    each node it self, when an array is passed with different rotations."""

    mesh_1 = create_test_mesh(get_default_test_beam_material)
    mesh_2 = create_test_mesh(get_default_test_beam_material)

    # Set the seed for the pseudo random numbers
    random.seed(0)

    # Rotate each node with a different rotation
    rotations = np.zeros([len(mesh_1.nodes), 4])
    origin = [100 * random.uniform(-1, 1) for _i in range(3)]
    for j, node in enumerate(mesh_1.nodes):
        rot = Rotation(
            [100 * random.uniform(-1, 1) for _i in range(3)],
            100 * random.uniform(-1, 1),
        )
        rotations[j, :] = rot.get_quaternion()
        node.rotate(rot, origin=origin)

    mesh_2.rotate(rotations, origin=origin)

    # Compare the output for the two meshes.
    assert_results_close(mesh_1, mesh_2)

    # Compare with reference results.
    assert_results_close(mesh_1, get_corresponding_reference_file_path())


@pytest.mark.parametrize("origin", [False, True])
@pytest.mark.parametrize("flip", [False, True])
def test_integration_core_mesh_reflection(
    origin,
    flip,
    get_default_test_beam_material,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Create a mesh, and its mirrored counterpart and then compare the input
    files."""

    # Rotations to be applied.
    rot_1 = Rotation([0, 1, 1], np.pi / 6)
    rot_2 = Rotation([1, 2.455, -1.2324], 1.2342352)

    mesh_ref = Mesh()
    mesh = Mesh()
    mat = get_default_test_beam_material(material_type="reissner")

    # Create the reference mesh.
    if not flip:
        create_beam_mesh_line(
            mesh_ref, Beam3rHerm2Line3, mat, [0, 0, 0], [1, 0, 0], n_el=1
        )
        create_beam_mesh_line(
            mesh_ref, Beam3rHerm2Line3, mat, [1, 0, 0], [1, 1, 0], n_el=1
        )
        create_beam_mesh_line(
            mesh_ref, Beam3rHerm2Line3, mat, [1, 1, 0], [1, 1, 1], n_el=1
        )
    else:
        create_beam_mesh_line(
            mesh_ref, Beam3rHerm2Line3, mat, [1, 0, 0], [0, 0, 0], n_el=1
        )
        create_beam_mesh_line(
            mesh_ref, Beam3rHerm2Line3, mat, [1, 1, 0], [1, 0, 0], n_el=1
        )
        create_beam_mesh_line(
            mesh_ref, Beam3rHerm2Line3, mat, [1, 1, 1], [1, 1, 0], n_el=1
        )

        # Reorder the internal nodes.
        old = mesh_ref.nodes.copy()
        mesh_ref.nodes[0] = old[2]
        mesh_ref.nodes[2] = old[0]
        mesh_ref.nodes[3] = old[5]
        mesh_ref.nodes[5] = old[3]
        mesh_ref.nodes[6] = old[8]
        mesh_ref.nodes[8] = old[6]

    mesh_ref.rotate(rot_1)

    # Create the mesh that will be mirrored.
    create_beam_mesh_line(mesh, Beam3rHerm2Line3, mat, [0, 0, 0], [-1, 0, 0], n_el=1)
    create_beam_mesh_line(mesh, Beam3rHerm2Line3, mat, [-1, 0, 0], [-1, 1, 0], n_el=1)
    create_beam_mesh_line(mesh, Beam3rHerm2Line3, mat, [-1, 1, 0], [-1, 1, 1], n_el=1)
    mesh.rotate(rot_1.inv())

    # Rotate everything, to show generalized reflection.
    mesh_ref.rotate(rot_2)
    mesh.rotate(rot_2)

    if origin:
        # Translate everything so the reflection plane is not in the
        # origin.
        r = [1, 2.455, -1.2324]
        mesh_ref.translate(r)
        mesh.translate(r)
        mesh.reflect(2 * (rot_2 * [1, 0, 0]), origin=r, flip_beams=flip)
    else:
        mesh.reflect(2 * (rot_2 * [1, 0, 0]), flip_beams=flip)

    # Compare the input files.
    assert_results_close(mesh_ref, mesh)

    # Compare with reference file.
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=("origin" if origin else "no_origin")
            + ("_flip" if flip else "_no_flip")
        ),
        mesh,
    )


@pytest.mark.parametrize(
    ("import_full", "radius", "reflect", "context"),
    [
        (False, None, True, nullcontext()),
        (False, 0.2, True, nullcontext()),
        (True, 0.2, False, nullcontext()),
        (False, 666, False, pytest.raises(ValueError)),
        (True, None, False, pytest.raises(ValueError)),
    ],
)
def test_integration_core_mesh_transformations_with_solid(
    import_full,
    radius,
    reflect,
    context,
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the different mesh transformation methods in combination with solid
    elements."""

    with context:
        # First, we create a line and wrap it with passing radius to the wrap function.

        # Create the mesh.
        input_file, mesh = import_four_c_model(
            input_file_path=get_corresponding_reference_file_path(
                reference_file_base_name="test_other_create_cubit_input_files_single_solid_element_brick"
            ),
            convert_input_to_mesh=import_full,
        )

        mat = get_default_test_beam_material(material_type="reissner")

        # Create the line.
        create_beam_mesh_line(
            mesh,
            Beam3rHerm2Line3,
            mat,
            [0.2, 0, 0],
            [0.2, 5 * 0.2 * 2 * np.pi, 4],
            n_el=3,
        )

        # Transform the mesh.
        mesh.wrap_around_cylinder(radius=radius)
        mesh.translate([1, 2, 3])
        mesh.rotate(Rotation([1, 2, 3], np.pi * 17.0 / 27.0))
        if reflect:
            mesh.reflect([0.1, -2, 1])

        input_file.add(mesh)

        # Check the output.
        assert_results_close(
            get_corresponding_reference_file_path(
                additional_identifier="full" if import_full else "yaml"
            ),
            input_file,
        )


def test_integration_core_mesh_wrap_cylinder_not_on_same_plane(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a helix that is itself wrapped around a cylinder."""

    # Ignore the warnings from wrap around cylinder.
    warnings.filterwarnings("ignore")

    # Create the mesh.
    mesh = Mesh()
    mat = get_default_test_beam_material(material_type="reissner")

    # Create the line and bend it to a helix.
    create_beam_mesh_line(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [0.2, 0, 0],
        [0.2, 5 * 0.2 * 2 * np.pi, 4],
        n_el=20,
    )
    mesh.wrap_around_cylinder()

    # Move the helix so its axis is in the y direction and goes through
    # (2 0 0). The helix is also moved by a lot in y-direction, this only
    # affects the angle phi when wrapping around a cylinder, not the shape
    # of the beam.
    mesh.rotate(Rotation([1, 0, 0], -0.5 * np.pi))
    mesh.translate([2, 666.666, 0])

    # Wrap the helix again.
    mesh.wrap_around_cylinder(radius=2.0)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_core_mesh_deep_copy(
    get_bc_data,
    get_default_test_beam_material,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """This test checks that the deep copy function on a mesh does not copy the
    materials or functions."""

    # Create material and function object.
    mat = get_default_test_beam_material(material_type="reissner")
    fun = Function([{"COMPONENT": 0, "SYMBOLIC_FUNCTION_OF_SPACE_TIME": "t"}])

    def create_mesh(mesh):
        """Add material and function to the mesh and create a beam."""
        mesh.add(fun, mat)
        set1 = create_beam_mesh_line(mesh, Beam3rHerm2Line3, mat, [0, 0, 0], [1, 0, 0])
        set2 = create_beam_mesh_line(mesh, Beam3rHerm2Line3, mat, [1, 0, 0], [1, 1, 0])
        mesh.add(
            BoundaryCondition(
                set1["line"], get_bc_data(identifier=1), bc_type=bme.bc.dirichlet
            )
        )
        mesh.add(
            BoundaryCondition(
                set2["line"], get_bc_data(identifier=2), bc_type=bme.bc.neumann
            )
        )
        mesh.couple_nodes()

    # The second mesh will be translated and rotated with those vales.
    translate = [1.0, 2.34535435, 3.345353]
    rotation = Rotation([1, 0.2342342423, -2.234234], np.pi / 15 * 27)

    # First create the mesh twice, move one and get the input file.
    mesh_ref_1 = Mesh()
    mesh_ref_2 = Mesh()
    create_mesh(mesh_ref_1)
    create_mesh(mesh_ref_2)
    mesh_ref_2.rotate(rotation)
    mesh_ref_2.translate(translate)

    mesh = Mesh()
    mesh.add(mesh_ref_1, mesh_ref_2)

    # Now copy the first mesh and add them together in the input file.
    mesh_copy_1 = Mesh()
    create_mesh(mesh_copy_1)
    mesh_copy_2 = mesh_copy_1.copy()
    mesh_copy_2.rotate(rotation)
    mesh_copy_2.translate(translate)

    mesh_copy = Mesh()
    mesh_copy.add(mesh_copy_1, mesh_copy_2)

    # Check that the input files are the same.
    assert_results_close(mesh, mesh_copy)
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_core_mesh_deep_copy_with_geometry_sets(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test that deep-copying a mesh together with geometry sets referencing it
    works such that the copied geometry sets also reference the copied mesh."""

    mesh = Mesh()
    beam_set = create_beam_mesh_line(
        mesh=mesh,
        beam_class=Beam3rHerm2Line3,
        material=get_default_test_beam_material(material_type="reissner"),
        start_point=[0, 0, 0],
        end_point=[1, 0, 0],
    )

    # Deep-copy both mesh and beam_set to keep node/element references consistent
    mesh_copy, beam_set_copy = copy.deepcopy((mesh, beam_set))

    mesh.add(mesh_copy)
    mesh.add(beam_set)
    mesh.add(beam_set_copy)
    bme.check_overlapping_elements = False
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_core_mesh_check_double_elements(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
    tmp_path,
):
    """Test the check for overlapping elements in a mesh."""

    # Create mesh object.
    mesh = Mesh()
    mat = get_default_test_beam_material()
    mesh.add(mat)

    # Add two beams to create an elbow structure. The beams each have a
    # node at the intersection.
    create_beam_mesh_line(mesh, Beam3, mat, [0, 0, 0], [2, 0, 0], n_el=2)
    create_beam_mesh_line(mesh, Beam3, mat, [0, 0, 0], [1, 0, 0])

    # Rotate the mesh with an arbitrary rotation.
    mesh.rotate(Rotation([1, 2, 3.24313], 2.2323423), [1, 3, -2.23232323])

    # The elements in the created mesh are overlapping, check that an error
    # is thrown.
    with pytest.raises(ValueError):
        mesh.check_overlapping_elements()

    # Check if the overlapping elements are written to the vtk output.
    warnings.filterwarnings("ignore")
    ref_file = get_corresponding_reference_file_path(
        additional_identifier="beam", extension="vtu"
    )
    vtk_file = tmp_path / "test_beam.vtu"
    mesh.write_vtk(
        output_name="test", output_directory=tmp_path, overlapping_elements=True
    )

    # Compare the vtk files.
    assert_results_close(ref_file, vtk_file)


def test_integration_core_mesh_display_pyvista(
    get_default_test_beam_material, get_corresponding_reference_file_path
):
    """Test that the display in pyvista function does not lead to errors.

    TODO: Add a check for the created visualization
    """

    _, mesh = create_beam_to_solid_conditions_model(
        get_default_test_beam_material,
        get_corresponding_reference_file_path,
        full_import=True,
    )

    mesh.display_pyvista(resolution=3)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module implements NURBS patches for the mesh."""

from abc import abstractmethod as _abstractmethod
from typing import Iterator as _Iterator

import numpy as _np

from beamme.core.conf import bme as _bme
from beamme.core.element import Element as _Element
from beamme.core.material import (
    MaterialSolidBase as _MaterialSolidBase,
)


class NURBSPatch(_Element):
    """A base class for a NURBS patch."""

    # A list of valid material types for this element
    valid_materials = [_MaterialSolidBase]

    def __init__(
        self,
        knot_vectors,
        polynomial_orders,
        material=None,
        nodes=None,
        data=None,
    ):
        super().__init__(nodes=nodes, material=material, data=data)

        # Knot vectors
        self.knot_vectors = knot_vectors

        # Polynomial degrees
        self.polynomial_orders = polynomial_orders

        # Set numbers for elements
        self.i_nurbs_patch = None

    def get_nurbs_dimension(self) -> int:
        """Determine the number of dimensions of the NURBS structure.

        Returns:
            Number of dimensions of the NURBS object.
        """
        n_knots = len(self.knot_vectors)
        n_polynomial = len(self.polynomial_orders)
        if not n_knots == n_polynomial:
            raise ValueError(
                "The variables n_knots and polynomial_orders should have "
                f"the same length. Got {n_knots} and {n_polynomial}"
            )
        return n_knots

    def get_number_of_control_points_per_dir(self) -> list[int]:
        """Determine the number of control points in each parameter direction
        of the patch.

        Returns:
            List of control points per direction.
        """
        n_dim = len(self.knot_vectors)
        n_cp_per_dim = []
        for i_dim in range(n_dim):
            knot_vector_size = len(self.knot_vectors[i_dim])
            polynomial_order = self.polynomial_orders[i_dim]
            n_cp_per_dim.append(knot_vector_size - polynomial_order - 1)
        return n_cp_per_dim

    def get_non_empty_knot_span_indices(self) -> list[list[int]]:
        """Determine the indices of the non-empty knot spans in each parameter
        direction.

        Returns:
            List of lists with the indices of the non-empty knot spans in
            each parameter direction.
        """

        non_empty_knot_spans_indices: list[list[int]] = [
            [] for _ in range(self.get_nurbs_dimension())
        ]

        for i_dir in range(len(self.knot_vectors)):
            for i_knot in range(len(self.knot_vectors[i_dir]) - 1):
                if (
                    abs(
                        self.knot_vectors[i_dir][i_knot]
                        - self.knot_vectors[i_dir][i_knot + 1]
                    )
                    > _bme.eps_knot_vector
                ):
                    non_empty_knot_spans_indices[i_dir].append(i_knot)
        return non_empty_knot_spans_indices

    def get_number_of_elements(self) -> int:
        """Determine the number of elements in this patch by checking the
        amount of nonzero knot spans in the knot vector.

        Returns:
            Number of elements for this patch.
        """

        non_empty_knot_spans_indices = self.get_non_empty_knot_span_indices()
        num_elements_dir = [len(indices) for indices in non_empty_knot_spans_indices]
        total_num_elements = _np.prod(num_elements_dir)
        return total_num_elements

    def _check_material(self) -> None:
        """Check if the linked material is valid for this type of NURBS solid
        element."""
        for material_type in type(self).valid_materials:
            if isinstance(self.material, material_type):
                return
        raise TypeError(
            f"NURBS solid of type {type(self)} can not have a material of"
            f" type {type(self.material)}!"
        )

    @_abstractmethod
    def get_knot_span_iterator(self) -> _Iterator[tuple[int, ...]]:
        """Return a tuple with the knot spans for this patch."""

    @_abstractmethod
    def get_ids_ctrlpts(self, *args) -> list[int]:
        """Compute the global indices of the control points that influence the
        element defined by the given knot span."""


class NURBSSurface(NURBSPatch):
    """A patch of a NURBS surface."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_knot_span_iterator(self) -> _Iterator[tuple[int, ...]]:
        """Return a tuple with the knot spans for this patch."""

        non_empty_knot_spans_indices = self.get_non_empty_knot_span_indices()
        return (
            (u, v)
            for v in non_empty_knot_spans_indices[1]
            for u in non_empty_knot_spans_indices[0]
        )

    def get_ids_ctrlpts(self, knot_span_u: int, knot_span_v: int) -> list[int]:
        """Compute the global indices of the control points that influence the
        element defined by the given knot span."""

        p, q = self.polynomial_orders
        ctrlpts_size_u = len(self.knot_vectors[0]) - p - 1
        id_u = knot_span_u - p
        id_v = knot_span_v - q

        return [
            ctrlpts_size_u * (id_v + j) + id_u + i
            for j in range(q + 1)
            for i in range(p + 1)
        ]


class NURBSVolume(NURBSPatch):
    """A patch of a NURBS volume."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_knot_span_iterator(self) -> _Iterator[tuple[int, ...]]:
        """Return a tuple with the knot spans for this patch."""

        non_empty_knot_spans_indices = self.get_non_empty_knot_span_indices()
        return (
            (u, v, w)
            for w in non_empty_knot_spans_indices[2]
            for v in non_empty_knot_spans_indices[1]
            for u in non_empty_knot_spans_indices[0]
        )

    def get_ids_ctrlpts(
        self, knot_span_u: int, knot_span_v: int, knot_span_w: int
    ) -> list[int]:
        """Compute the global indices of the control points that influence the
        element defined by the given knot span."""

        p, q, r = self.polynomial_orders
        id_u = knot_span_u - p
        id_v = knot_span_v - q
        id_w = knot_span_w - r
        size_u = len(self.knot_vectors[0]) - p - 1
        size_v = len(self.knot_vectors[1]) - q - 1

        return [
            size_u * size_v * (id_w + k) + size_u * (id_v + j) + id_u + i
            for k in range(r + 1)
            for j in range(q + 1)
            for i in range(p + 1)
        ]



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Unit tests for the dbc monitor."""

import numpy as np

from beamme.four_c.dbc_monitor import linear_time_transformation, read_dbc_monitor_file


def test_beamme_four_c_dbc_monitor_read_dbc_monitor_file(
    assert_results_close, get_corresponding_reference_file_path
):
    """Test that a dbc monitor file can be read correctly."""

    nodes, time, force, moment = read_dbc_monitor_file(
        get_corresponding_reference_file_path(extension="yaml")
    )
    assert_results_close(nodes, [2, 4, 5, 9, 10])
    assert_results_close(time, [0.0, 0.1, 0.2])
    assert_results_close(force, [[0, 0, 0], [0.1, 0.2, 0.3], [0.7, 0.8, 0.9]])
    assert_results_close(moment, [[0, 0, 0], [0.4, 0.5, 0.6], [1.0, 1.1, 1.2]])


def test_beamme_four_c_dbc_monitor_linear_time_transformation_scaling():
    """Test the scaling of the interval for the function.

    Starts with a function within the interval [0,1] and transforms
    them.
    """

    # starting time array
    time = np.array([0, 0.5, 0.75, 1.0])

    # corresponding values 3 values per time step
    force = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])

    # with the result vector
    force_result = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])

    # base case no scaling only
    time_trans, force_trans = linear_time_transformation(
        time, force, [0, 1], flip=False
    )

    # first result is simply the attached
    time_result = np.array([0, 0.5, 0.75, 1.0])

    # check solution
    assert time_trans.tolist() == time_result.tolist()
    assert force_trans.tolist() == force_result.tolist()

    # transform to interval [0, 2]
    time_trans, force_trans = linear_time_transformation(
        time, force, [0, 2], flip=False
    )

    # time values should double
    assert time_trans.tolist() == (2 * time_result).tolist()
    assert force_trans.tolist() == force_result.tolist()

    # new result
    force_result = np.array(
        [[1, 2, 3], [1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12], [10, 11, 12]]
    )

    # shift to the interval [1 ,2] and add valid start end point
    time_trans, force_trans = linear_time_transformation(
        time, force, [1, 2, 5], flip=False, valid_start_and_end_point=True
    )
    assert time_trans.tolist() == np.array([0, 1.0, 1.5, 1.75, 2.0, 5.0]).tolist()
    assert force_trans.tolist() == force_result.tolist()


def test_beamme_four_c_dbc_monitor_linear_time_transformation_flip():
    """Test the flip flag option of linear_time_transformation to mirror the
    function."""

    # base case no scaling no end points should be attached
    # starting time array
    time = np.array([0, 0.5, 0.75, 1.0])

    # corresponding values:  3 values per time step
    force = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])

    # first result is simply the attached point at the end
    time_result = np.array([0, 0.25, 0.5, 1.0])

    # with the value vector:
    force_result = np.array([[10, 11, 12], [7, 8, 9], [4, 5, 6], [1, 2, 3]])

    # base case no scaling only end points should be attached
    time_trans, force_trans = linear_time_transformation(time, force, [0, 1], flip=True)

    # check solution
    assert time_result.tolist() == time_trans.tolist()
    assert force_trans.tolist() == force_result.tolist()

    # new force result
    force_result = np.array([[10, 11, 12], [7, 8, 9], [4, 5, 6], [1, 2, 3]])

    time_result = np.array([0, 0.25, 0.5, 1.0]) + 1

    # test now an shift to the interval [1 ,2]
    time_trans, force_trans = linear_time_transformation(time, force, [1, 2], flip=True)
    assert time_result.tolist() == time_trans.tolist()
    assert force_trans.tolist() == force_result.tolist()

    # same trick as above but with 2
    time_result = np.array([0, 2.0, 2.25, 2.5, 3.0, 5.0])
    # new force result
    force_result = np.array(
        [[10, 11, 12], [10, 11, 12], [7, 8, 9], [4, 5, 6], [1, 2, 3], [1, 2, 3]]
    )

    # test offset and scaling and add valid start and end point
    time_trans, force_trans = linear_time_transformation(
        time, force, [2, 3, 5], flip=True, valid_start_and_end_point=True
    )
    assert time_result.tolist() == time_trans.tolist()
    assert force_trans.tolist() == force_result.tolist()



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Integration tests for materials in 4C."""

import pytest

from beamme.core.mesh import Mesh
from beamme.four_c.input_file import InputFile
from beamme.four_c.material import MaterialSolid


def test_integration_four_c_sub_materials(
    get_default_test_solid_material,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Check if sub-materials are handled correctly."""

    # Add a nested material to the mesh and check the result.
    mesh = Mesh()
    material = get_default_test_solid_material(material_type="solid_nested")
    mesh.add(material)
    assert_results_close(get_corresponding_reference_file_path(), mesh)

    # Add the material again and check that the result is the same.
    mesh.add(material)
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_four_c_sub_materials_error():
    """Check the error for incorrectly added sub-materials."""

    mesh = Mesh()
    material_sub = MaterialSolid(
        material_string="ELAST_CoupSVK", data={"YOUNG": 1.0, "NUE": 0.0}
    )
    mesh.add(material_sub)
    material = MaterialSolid(
        material_string="MAT_ElastHyper",
        data={
            "NUMMAT": 1,
            "MATIDS": [material_sub],
            "DENS": 1.0,
        },
    )
    mesh.add(material)
    input_file = InputFile()

    with pytest.raises(
        ValueError,
        match="Materials are not unique!",
    ):
        input_file.add(mesh)


def test_integration_four_c_sub_materials_material_numbering(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test that materials can be added as structured data (dictionaries) to an
    input file (as is done when importing input files) and that the numbering
    with other added materials does not lead to materials with double IDs."""

    input_file = InputFile()
    input_file.add(
        {
            "MATERIALS": [
                {
                    "MAT": 1,
                    "MAT_ViscoElastHyper": {
                        "NUMMAT": 4,
                        "MATIDS": [10, 11, 12, 13],
                        "DENS": 1.3e-06,
                    },
                },
                {
                    "MAT": 10,
                    "ELAST_CoupNeoHooke": {
                        "YOUNG": 0.16,
                        "NUE": 0.45,
                    },
                },
                {
                    "MAT": 11,
                    "VISCO_GenMax": {
                        "TAU": 0.1,
                        "BETA": 0.4,
                        "SOLVE": "OST",
                    },
                },
                {
                    "MAT": 12,
                    "ELAST_CoupAnisoExpo": {
                        "K1": 0.0024,
                        "K2": 0.14,
                        "GAMMA": 0,
                        "K1COMP": 0,
                        "K2COMP": 1,
                        "STR_TENS_ID": 100,
                        "INIT": 3,
                    },
                },
                {
                    "MAT": 13,
                    "ELAST_CoupAnisoExpo": {
                        "K1": 0.0054,
                        "K2": 1.24,
                        "GAMMA": 0,
                        "K1COMP": 0,
                        "K2COMP": 1,
                        "STR_TENS_ID": 100,
                        "INIT": 3,
                        "FIBER_ID": 2,
                    },
                },
                {
                    "MAT": 100,
                    "ELAST_StructuralTensor": {
                        "STRATEGY": "Standard",
                    },
                },
                {
                    "MAT": 2,
                    "MAT_ElastHyper": {
                        "NUMMAT": 3,
                        "MATIDS": [20, 21, 22],
                        "DENS": 1.3e-06,
                    },
                },
                {
                    "MAT": 20,
                    "ELAST_CoupNeoHooke": {
                        "YOUNG": 1.23,
                        "NUE": 0.45,
                    },
                },
                {
                    "MAT": 21,
                    "ELAST_CoupAnisoExpo": {
                        "K1": 0.0004,
                        "K2": 12,
                        "GAMMA": 0,
                        "K1COMP": 0,
                        "K2COMP": 1,
                        "STR_TENS_ID": 200,
                        "INIT": 3,
                    },
                },
                {
                    "MAT": 22,
                    "ELAST_CoupAnisoExpo": {
                        "K1": 0.0502,
                        "K2": 10,
                        "GAMMA": 0,
                        "K1COMP": 0,
                        "K2COMP": 1,
                        "STR_TENS_ID": 200,
                        "INIT": 3,
                        "FIBER_ID": 2,
                    },
                },
                {
                    "MAT": 200,
                    "ELAST_StructuralTensor": {
                        "STRATEGY": "Standard",
                    },
                },
            ]
        }
    )

    mesh = Mesh()
    mesh.add(get_default_test_beam_material(material_type="reissner"))

    input_file.add(mesh)

    assert_results_close(get_corresponding_reference_file_path(), input_file)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test general functionality of the BeamMe Abaqus
module with end-to-end integration tests."""

import numpy as np
import pytest

from beamme.abaqus.beam import generate_abaqus_beam
from beamme.abaqus.input_file import AbaqusBeamNormalDefinition, AbaqusInputFile
from beamme.abaqus.material import AbaqusBeamMaterial
from beamme.core.geometry_set import GeometrySet
from beamme.core.mesh import Mesh
from beamme.core.rotation import Rotation
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line

PYTEST_ABAQUS_NORMAL_DEFINITION_PARAMETRIZE = [
    ("normal_definition", "additional_identifier"),
    [
        (AbaqusBeamNormalDefinition.normal, "normal"),
        (AbaqusBeamNormalDefinition.normal_and_extra_node, "normal_and_extra_node"),
    ],
]


@pytest.mark.parametrize(*PYTEST_ABAQUS_NORMAL_DEFINITION_PARAMETRIZE)
def test_integration_abaqus_helix(
    normal_definition,
    additional_identifier,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a helix and check the created Abaqus input file."""

    # Helix parameters
    n_el = 10
    turns = 1.5
    length = 1.0
    r = 0.5

    mesh = Mesh()
    mat = AbaqusBeamMaterial("beam_material")
    beam_type = generate_abaqus_beam("B32H")
    helix_set = create_beam_mesh_line(
        mesh,
        beam_type,
        mat,
        [r, 0, 0],
        [r, r * 2.0 * np.pi * turns, length],
        n_el=n_el,
    )
    mesh.wrap_around_cylinder()

    start_set = helix_set["start"]
    start_set.name = "fix_node"
    mesh.add(start_set)

    end_set = helix_set["end"]
    end_set.name = "load_node"
    mesh.add(end_set)

    end_set = helix_set["line"]
    end_set.name = "beam_elements"
    mesh.add(end_set)

    input_file = AbaqusInputFile(mesh)
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier, extension="inp"
        ),
        input_file.get_input_file_string(normal_definition),
        atol=1e-14,
    )


@pytest.mark.parametrize(*PYTEST_ABAQUS_NORMAL_DEFINITION_PARAMETRIZE)
def test_integration_abaqus_frame(
    normal_definition,
    additional_identifier,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a frame out of connected beams with different materials."""

    mesh = Mesh()
    mat_1 = AbaqusBeamMaterial("beam_material_1")
    mat_2 = AbaqusBeamMaterial("beam_material_2")
    beam_type_b23 = generate_abaqus_beam("B32H")
    beam_type_b33 = generate_abaqus_beam("B33H")

    create_beam_mesh_line(mesh, beam_type_b23, mat_1, [0, 0, 0], [1, 0, 0], n_el=2)
    mesh.rotate(Rotation([1, 0, 0], np.pi * 0.5))
    create_beam_mesh_line(mesh, beam_type_b23, mat_2, [1, 0, 0], [1, 1, 0], n_el=2)
    create_beam_mesh_line(mesh, beam_type_b33, mat_1, [1, 1, 0], [1, 1, 1], n_el=2)
    create_beam_mesh_line(mesh, beam_type_b33, mat_2, [1, 1, 1], [0, 1, 1], n_el=2)
    mesh.couple_nodes()

    fix_set = GeometrySet(mesh.nodes[0], name="fix_node")
    load_set = GeometrySet(mesh.nodes[-1], name="load_node")
    mesh.add(fix_set, load_set)

    input_file = AbaqusInputFile(mesh)
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier, extension="inp"
        ),
        input_file.get_input_file_string(normal_definition),
        atol=1e-15,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test the performance of the geometric space time module."""

import pytest

from beamme.core.element_beam import generate_beam_class
from beamme.core.material import MaterialBeamBase
from beamme.core.mesh import Mesh
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line
from beamme.space_time.beam_to_space_time import beam_to_space_time


@pytest.mark.performance
def test_performance_space_time_create_mesh_in_space(
    evaluate_execution_time, cache_data
):
    """Test the performance of the mesh creation in space."""

    mesh = Mesh()
    beam_type = generate_beam_class(3)

    evaluate_execution_time(
        "BeamMe: Space-Time: Create mesh in space",
        create_beam_mesh_line,
        kwargs={
            "mesh": mesh,
            "beam_class": beam_type,
            "material": MaterialBeamBase(),
            "start_point": [0, 0, 0],
            "end_point": [1, 0, 0],
            "n_el": 100,
        },
        expected_time=0.01,
    )

    # store mesh in cache for upcoming test
    cache_data.mesh = mesh


@pytest.mark.performance
def test_performance_space_time_create_mesh_in_time(
    evaluate_execution_time, cache_data
):
    """Test the performance of the mesh creation in time."""

    evaluate_execution_time(
        "BeamMe: Space-Time: Create mesh in time",
        beam_to_space_time,
        kwargs={
            "mesh_space_or_generator": cache_data.mesh,
            "time_duration": 6.9,
            "number_of_elements_in_time": 1000,
            "time_start": 1.69,
        },
        expected_time=3.0,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Integration tests for the solid shell capability of BeamMe."""

import numpy as np

from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.four_c.model_importer import import_four_c_model
from beamme.four_c.solid_shell_thickness_direction import (
    get_visualization_third_parameter_direction_hex8,
    set_solid_shell_thickness_direction,
)
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line


def test_integration_four_c_solid_shell_thickness_direction_block(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the solid shell direction detection functionality for a block
    geometry."""

    # Test the plates
    _, mesh_block = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            reference_file_base_name="test_other_create_cubit_input_files_solid_shell_blocks"
        ),
        convert_input_to_mesh=True,
    )

    # Add a beam element to check the function also works with beam elements
    create_beam_mesh_line(
        mesh_block,
        Beam3rHerm2Line3,
        get_default_test_beam_material(material_type="reissner"),
        [0, 0, 0],
        [1, 0, 0],
        n_el=1,
    )
    # Set the thickness direction and compare result
    set_solid_shell_thickness_direction(mesh_block.elements, selection_type="thickness")
    assert_results_close(get_corresponding_reference_file_path(), mesh_block)


def test_integration_four_c_solid_shell_thickness_direction_dome(
    assert_results_close, get_corresponding_reference_file_path
):
    """Test the solid shell direction detection functionality for a dome
    geometry."""

    _, mesh_dome_original = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            reference_file_base_name="test_other_create_cubit_input_files_solid_shell_dome"
        ),
        convert_input_to_mesh=True,
    )

    # Test that the thickness version works
    mesh_dome = mesh_dome_original.copy()
    set_solid_shell_thickness_direction(mesh_dome.elements, selection_type="thickness")
    assert_results_close(
        get_corresponding_reference_file_path(additional_identifier="thickness"),
        mesh_dome,
    )

    # Test that the direction function version works (same result as thickness)
    def director_function(cell_center):
        """Return director that will be used to determine the solid thickness
        direction."""
        return cell_center / np.linalg.norm(cell_center)

    mesh_dome = mesh_dome_original.copy()
    set_solid_shell_thickness_direction(
        mesh_dome.elements,
        selection_type="projection_director_function",
        director_function=director_function,
    )
    assert_results_close(
        get_corresponding_reference_file_path(additional_identifier="thickness"),
        mesh_dome,
    )

    # Test that the constant direction version works
    mesh_dome = mesh_dome_original.copy()
    set_solid_shell_thickness_direction(
        mesh_dome.elements,
        selection_type="projection_director",
        director=[0, 0, 1],
        identify_threshold=None,
    )
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier="constant_direction"
        ),
        mesh_dome,
    )


def test_integration_four_c_solid_shell_thickness_direction_visualization(
    assert_results_close, get_corresponding_reference_file_path, tmp_path
):
    """Test the solid shell direction visualization functionality."""

    _, mesh_dome = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            reference_file_base_name="test_other_create_cubit_input_files_solid_shell_dome"
        ),
        convert_input_to_mesh=True,
    )

    ref_file = get_corresponding_reference_file_path(extension="vtu")
    test_file = tmp_path / ref_file.name

    set_solid_shell_thickness_direction(
        mesh_dome.elements,
        selection_type="projection_director",
        director=[0, 0, 1],
        identify_threshold=None,
    )
    grid = get_visualization_third_parameter_direction_hex8(mesh_dome)
    grid.save(test_file)
    assert_results_close(ref_file, test_file)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Helper functions to interact with the BeamMe environment."""

import os as _os
import shutil as _shutil
import subprocess as _subprocess  # nosec B404
import sys as _sys
from importlib.util import find_spec as _find_spec
from pathlib import Path as _Path
from typing import Optional as _Optional
from typing import Tuple as _Tuple


def cubitpy_is_available() -> bool:
    """Check if CubitPy is installed.

    Returns:
        True if CubitPy is installed, False otherwise
    """

    if _find_spec("cubitpy") is None:
        return False
    return True


def is_mybinder():
    """Check if the current environment is running on mybinder."""
    return "BINDER_LAUNCH_HOST" in _os.environ.keys()


def is_testing():
    """Check if the current environment is a pytest testing run."""
    return "PYTEST_CURRENT_TEST" in _os.environ


def get_env_variable(name, *, default="default_not_set"):
    """Return the value of an environment variable.

    Args
    ----
    name: str
        Name of the environment variable
    default:
        Value to be returned if the given named environment variable does
        not exist. If this is not set and the name is not in the env
        variables, then an error will be thrown.
    """
    if name in _os.environ.keys():
        return _os.environ[name]
    elif default == "default_not_set":
        raise ValueError(f"Environment variable {name} is not set")
    return default


def get_git_data(repo_path: _Path) -> _Tuple[_Optional[str], _Optional[str]]:
    """Return the hash and date of the current git commit of a git repo for a
    given repo path.

    Args:
        repo_path: Path to the git repository.
    Returns:
        A tuple with the hash and date of the current git commit
        if available, otherwise None.
    """

    git = _shutil.which("git")
    if git is None:
        raise RuntimeError("Git executable not found")

    out_sha = _subprocess.run(  # nosec B603
        [git, "rev-parse", "HEAD"],
        cwd=repo_path,
        stdout=_subprocess.PIPE,
        stderr=_subprocess.DEVNULL,
    )

    out_date = _subprocess.run(  # nosec B603
        [git, "show", "-s", "--format=%ci"],
        cwd=repo_path,
        stdout=_subprocess.PIPE,
        stderr=_subprocess.DEVNULL,
    )

    if not out_sha.returncode + out_date.returncode == 0:
        return None, None

    git_sha = out_sha.stdout.decode("ascii").strip()
    git_date = out_date.stdout.decode("ascii").strip()

    return git_sha, git_date


def get_application_path() -> _Path | None:
    """Returns the application path which created this input file and ensures
    that the file exists.

    Returns:
        A path to the file, which created this input file or None.
    """

    # return valid application path if it exists.
    if _Path(_sys.argv[0]).resolve().exists():
        return _Path(_sys.argv[0]).resolve()

    return None



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module implements the class that represents one node in the Mesh."""

import numpy as _np

from beamme.core.base_mesh_item import BaseMeshItem as _BaseMeshItem
from beamme.core.rotation import Rotation as _Rotation


class Node(_BaseMeshItem):
    """This object represents one node in the mesh."""

    def __init__(self, coordinates, *, is_middle_node=False, **kwargs):
        super().__init__(**kwargs)

        # Coordinates of this node.
        self.coordinates = _np.array(coordinates)

        # If this node is at the end of a line or curve (by default only those
        # nodes are checked for overlapping nodes).
        self.is_end_node = False

        # If the node is in the middle of a beam element.
        self.is_middle_node = is_middle_node

        # Lists with the objects that this node is linked to.
        self.element_link = []
        self.node_sets_link = []
        self.element_partner_index = None
        self.mesh = None

        # If this node is replaced, store a link to the remaining node.
        self.master_node = None

    def get_master_node(self):
        """Return the master node of this node.

        If the node has not been replaced, this object is returned.
        """

        if self.master_node is None:
            return self
        else:
            return self.master_node.get_master_node()

    def replace_with(self, master_node):
        """Replace this node with another node object."""

        # Check that the two nodes have the same type.
        if not isinstance(self, type(master_node)):
            raise TypeError(
                "A node can only be replaced by a node with the same type. "
                + f"Got {type(self)} and {type(master_node)}"
            )

        # Replace the links to this node in the referenced objects.
        self.mesh.replace_node(self, master_node)
        for element in self.element_link:
            element.replace_node(self, master_node)
        for node_set in self.node_sets_link:
            node_set.replace_node(self, master_node)

        # Set link to master node.
        self.master_node = master_node.get_master_node()

    def unlink(self):
        """Reset the links to elements, node sets and global indices."""
        self.element_link = []
        self.node_sets_link = []
        self.mesh = None
        self.i_global = None

    def rotate(self, *args, **kwargs):
        """Don't do anything for a standard node, as this node can not be
        rotated."""


class NodeCosserat(Node):
    """This object represents a Cosserat node in the mesh, i.e., it contains
    three positions and three rotations."""

    def __init__(
        self,
        coordinates,
        rotation: _Rotation,
        *,
        arc_length: float | None = None,
        **kwargs,
    ):
        super().__init__(coordinates, **kwargs)

        # Rotation of this node.
        self.rotation = rotation.copy()

        # Arc length along the filament that this beam is a part of
        self.arc_length = arc_length

    def rotate(self, rotation, *, origin=None, only_rotate_triads=False):
        """Rotate this node.

        By default the node is rotated around the origin (0,0,0), if the
        keyword argument origin is given, it is rotated around that
        point. If only_rotate_triads is True, then only the rotation is
        affected, the position of the node stays the same.
        """

        self.rotation = rotation * self.rotation

        # Rotate the positions (around origin).
        if not only_rotate_triads:
            if origin is not None:
                self.coordinates = self.coordinates - origin
            self.coordinates = rotation * self.coordinates
            if origin is not None:
                self.coordinates = self.coordinates + origin


class ControlPoint(Node):
    """This object represents a control point with a weight in the mesh."""

    def __init__(self, coordinates, weight, **kwargs):
        super().__init__(coordinates, **kwargs)

        # Weight of this node
        self.weight = weight



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test environment utils."""

from unittest.mock import patch

from beamme.geometric_search.utils import (
    arborx_is_available,
    cython_is_available,
)


def test_beamme_geometric_search_utils_is_arborx_available() -> None:
    """Test is_arborx_available function."""

    with patch("beamme.geometric_search.utils._find_spec", return_value=True):
        assert arborx_is_available() is True

    with patch("beamme.geometric_search.utils._find_spec", return_value=None):
        assert arborx_is_available() is False


def test_beamme_geometric_search_utils_is_cython_available() -> None:
    """Test is_cython_available function."""

    with patch("beamme.geometric_search.utils._find_spec", return_value=True):
        assert cython_is_available() is True

    with patch("beamme.geometric_search.utils._find_spec", return_value=None):
        assert cython_is_available() is False



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Generic function for beam creation."""

from typing import Callable as _Callable
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Type as _Type
from typing import Union as _Union

import numpy as _np

from beamme.core.conf import bme as _bme
from beamme.core.element_beam import Beam as _Beam
from beamme.core.geometry_set import GeometryName as _GeometryName
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.material import MaterialBeamBase as _MaterialBeamBase
from beamme.core.mesh import Mesh as _Mesh
from beamme.core.node import NodeCosserat as _NodeCosserat
from beamme.utils.nodes import get_single_node as _get_single_node


def create_beam_mesh_generic(
    mesh: _Mesh,
    *,
    beam_class: _Type[_Beam],
    material: _MaterialBeamBase,
    function_generator: _Callable,
    interval: _Tuple[float, float],
    n_el: _Optional[int] = None,
    l_el: _Optional[float] = None,
    interval_length: _Optional[float] = None,
    set_nodal_arc_length: bool = False,
    nodal_arc_length_offset: _Optional[float] = None,
    node_positions_of_elements: _Optional[_List[float]] = None,
    start_node: _Optional[_Union[_NodeCosserat, _GeometrySet]] = None,
    end_node: _Optional[_Union[_NodeCosserat, _GeometrySet]] = None,
    close_beam: bool = False,
    vtk_cell_data: _Optional[_Dict[str, _Tuple]] = None,
) -> _GeometryName:
    """Generic beam creation function.

    Remark for given start and/or end nodes:
        If the rotation does not match, but the tangent vector is the same,
        the created beams triads are rotated so the physical problem stays
        the same (for axi-symmetric beam cross-sections) but the nodes can
        be reused.

    Args:
        mesh:
            Mesh that the created beam(s) should be added to.
        beam_class:
            Class of beam that will be used for this line.
        material:
            Material for this line.
        function_generator:
            The function_generator has to take two variables, point_a and
            point_b (both within the interval) and return a function(xi) that
            calculates the position and rotation along the beam, where
            point_a -> xi = -1 and point_b -> xi = 1.

            Usually, the Jacobian of the returned position field should be a unit
            vector. Otherwise, the nodes may be spaced in an undesired way. All
            standard mesh creation functions fulfill this property.
        interval:
            Start and end values for interval that will be used to create the
            beam.
        n_el:
            Number of equally spaced beam elements along the line. Defaults to 1.
            Mutually exclusive with l_el
        l_el:
            Desired length of beam elements. This requires the option interval_length
            to be set. Mutually exclusive with n_el. Be aware, that this length
            might not be achieved, if the elements are warped after they are
            created.
        interval_length:
            Total length of the interval. Is required when the option l_el is given.
        set_nodal_arc_length:
            Flag if the arc length along the beam filament is set in the created
            nodes. It is ensured that the arc length is consistent with possible
            given start/end nodes.
        nodal_arc_length_offset:
            Offset of the stored nodal arc length w.r.t. to the one generated by
            the function. Defaults to 0, the arc length set in the start node, or
            the arc length in the end node minus total length (such that the arc
            length at the end node matches).
        node_positions_of_elements:
            A list of normalized positions (within [0,1] and in ascending order)
            that define the boundaries of beam elements along the created curve.
            The given values will be mapped to the actual `interval` given as an
            argument to this function. These values specify where elements start
            and end, additional internal nodes (such as midpoints in higher-order
            elements) may be placed automatically.
        start_node:
            Node to use as the first node for this line. Use this if the line
            is connected to other lines (angles have to be the same, otherwise
            connections should be used). If a geometry set is given, it can
            contain one, and one node only.
        end_node:
            If this is a Node or GeometrySet, the last node of the created beam
            is set to that node.
            If it is True the created beam is closed within itself.
        close_beam:
            If it is True the created beam is closed within itself (mutually
            exclusive with end_node).
        vtk_cell_data:
            With this argument, a vtk cell data can be set for the elements
            created within this function. This can be used to check which
            elements are created by which function.

    Returns:
        Geometry sets with the 'start' and 'end' node of the curve. Also a 'line' set
        with all nodes of the curve.
    """

    # Check for mutually exclusive parameters
    n_given_arguments = sum(
        1
        for argument in [n_el, l_el, node_positions_of_elements]
        if argument is not None
    )
    if n_given_arguments == 0:
        # No arguments were given, use a single element per default
        n_el = 1
    elif n_given_arguments > 1:
        raise ValueError(
            'The arguments "n_el", "l_el" and "node_positions_of_elements" are mutually exclusive'
        )

    if close_beam and end_node is not None:
        raise ValueError(
            'The arguments "close_beam" and "end_node" are mutually exclusive'
        )

    if set_nodal_arc_length:
        if close_beam:
            raise ValueError(
                "The flags 'set_nodal_arc_length' and 'close_beam' are mutually exclusive."
            )
    elif nodal_arc_length_offset is not None:
        raise ValueError(
            'Providing the argument "nodal_arc_length_offset" without setting '
            '"set_nodal_arc_length" to True does not make sense.'
        )

    # Cases where we have equally spaced elements
    if n_el is not None or l_el is not None:
        if l_el is not None:
            # Calculate the number of elements in case a desired element length is provided
            if interval_length is None:
                raise ValueError(
                    'The parameter "l_el" requires "interval_length" to be set.'
                )
            n_el = max([1, round(interval_length / l_el)])
        elif n_el is None:
            raise ValueError("n_el should not be None at this point")

        node_positions_of_elements = [i_node / n_el for i_node in range(n_el + 1)]
    # A list for the element node positions was provided
    elif node_positions_of_elements is not None:
        # Check that the given positions are in ascending order and start with 0 and end with 1
        for index, value, name in zip([0, -1], [0, 1], ["First", "Last"]):
            if not _np.isclose(
                value,
                node_positions_of_elements[index],
                atol=1e-12,
                rtol=0.0,
            ):
                raise ValueError(
                    f"{name} entry of node_positions_of_elements must be {value}, got {node_positions_of_elements[index]}"
                )
        if not all(
            x < y
            for x, y in zip(node_positions_of_elements, node_positions_of_elements[1:])
        ):
            raise ValueError(
                f"The given node_positions_of_elements must be in ascending order. Got {node_positions_of_elements}"
            )

    # Get the scale the node positions to the interval coordinates
    interval_node_positions_of_elements = interval[0] + (
        interval[1] - interval[0]
    ) * _np.asarray(node_positions_of_elements)

    # We need to make sure we have the number of elements for the case a given end node
    n_el = len(interval_node_positions_of_elements) - 1

    # Make sure the material is in the mesh.
    mesh.add_material(material)

    # List with nodes and elements that will be added in the creation of
    # this beam.
    elements = []
    nodes = []

    def check_given_node(node):
        """Check that the given node is already in the mesh."""
        if node not in mesh.nodes:
            raise ValueError("The given node is not in the current mesh")

    def get_relative_twist(rotation_node, rotation_function, name):
        """Check if the rotation at a node and the one returned by the function
        match.

        If not, check if the first basis vector of the triads is the
        same. If that is the case, a simple relative twist can be
        applied to ensure that the triad field is continuous. This
        relative twist can lead to issues if the beam cross-section is
        not double symmetric.
        """

        if rotation_node == rotation_function:
            return None
        elif not _bme.allow_beam_rotation:
            # The settings do not allow for a rotation of the beam
            raise ValueError(
                f"Given rotation of the {name} node does not match with given rotation function!"
            )
        else:
            # Evaluate the relative rotation
            # First check if the first basis vector is the same
            relative_basis_1 = rotation_node.inv() * rotation_function * [1, 0, 0]
            if _np.linalg.norm(relative_basis_1 - [1, 0, 0]) < _bme.eps_quaternion:
                # Calculate the relative rotation
                return rotation_function.inv() * rotation_node
            else:
                raise ValueError(
                    f"The tangent of the {name} node does not match with the given function!"
                )

    # Position and rotation at the start and end of the interval
    function_over_whole_interval = function_generator(*interval)
    relative_twist_start = None
    relative_twist_end = None

    # If a start node is given, set this as the first node for this beam.
    if start_node is not None:
        start_node = _get_single_node(start_node)
        nodes = [start_node]
        check_given_node(start_node)
        _, start_rotation, _ = function_over_whole_interval(-1.0)
        relative_twist_start = get_relative_twist(
            start_node.rotation, start_rotation, "start"
        )

    # If an end node is given, check what behavior is wanted.
    if end_node is not None:
        end_node = _get_single_node(end_node)
        check_given_node(end_node)
        _, end_rotation, _ = function_over_whole_interval(1.0)
        relative_twist_end = get_relative_twist(end_node.rotation, end_rotation, "end")

    # Get the start value for the arc length functionality
    if set_nodal_arc_length:
        if nodal_arc_length_offset is not None:
            # Let's use the given value, if it does not match with possible given
            # start or end nodes, the check in the create beam function will detect
            # that.
            pass
        elif start_node is not None and start_node.arc_length is not None:
            nodal_arc_length_offset = start_node.arc_length
        elif end_node is not None and end_node.arc_length is not None:
            if interval_length is None:
                raise ValueError(
                    "The end node has an arc length but the interval_length is not "
                    "given. This is not supported."
                )
            nodal_arc_length_offset = end_node.arc_length - interval_length
        else:
            # Default value
            nodal_arc_length_offset = 0.0

    # Check if a relative twist has to be applied
    if relative_twist_start is not None and relative_twist_end is not None:
        if relative_twist_start == relative_twist_end:
            relative_twist = relative_twist_start
        else:
            raise ValueError(
                "The relative twist required for the start and end node do not match"
            )
    elif relative_twist_start is not None:
        relative_twist = relative_twist_start
    elif relative_twist_end is not None:
        relative_twist = relative_twist_end
    else:
        relative_twist = None

    # Create the beams.
    for i_el in range(n_el):
        # If the beam is closed with itself, set the end node to be the
        # first node of the beam. This is done when the second element is
        # created, as the first node already exists here.
        if i_el == 1 and close_beam:
            end_node = nodes[0]

        # Get the function to create this beam element.
        function = function_generator(
            interval_node_positions_of_elements[i_el],
            interval_node_positions_of_elements[i_el + 1],
        )

        # Set the start node for the created beam.
        if start_node is not None or i_el > 0:
            first_node = nodes[-1]
        else:
            first_node = None

        # If an end node is given, set this one for the last element.
        if end_node is not None and i_el == n_el - 1:
            last_node = end_node
        else:
            last_node = None

        element, created_nodes = beam_class.create_beam(
            material,
            function,
            start_node=first_node,
            end_node=last_node,
            relative_twist=relative_twist,
            set_nodal_arc_length=set_nodal_arc_length,
            nodal_arc_length_offset=nodal_arc_length_offset,
        )
        elements.append(element)
        nodes.extend(created_nodes)

    # Set vtk cell data on created elements.
    if vtk_cell_data is not None:
        for data_name, data_value in vtk_cell_data.items():
            for element in elements:
                if data_name in element.vtk_cell_data.keys():
                    raise KeyError(
                        'The cell data "{}" already exists!'.format(data_name)
                    )
                element.vtk_cell_data[data_name] = data_value

    # Add items to the mesh
    mesh.elements.extend(elements)
    if start_node is None:
        mesh.nodes.extend(nodes)
    else:
        mesh.nodes.extend(nodes[1:])

    # Set the last node of the beam.
    if end_node is None:
        end_node = nodes[-1]

    # Set the nodes that are at the beginning and end of line (for search
    # of overlapping points)
    nodes[0].is_end_node = True
    end_node.is_end_node = True

    # Create geometry sets that will be returned.
    return_set = _GeometryName()
    return_set["start"] = _GeometrySet(nodes[0])
    return_set["end"] = _GeometrySet(end_node)
    return_set["line"] = _GeometrySet(elements)

    return return_set



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Provide a function that allows to run 4C."""

import os as _os
import shutil as _shutil
import subprocess as _subprocess  # nosec B404
import sys as _sys
from pathlib import Path as _Path

from beamme.utils.environment import get_env_variable as _get_env_variable


def run_four_c(
    input_file,
    output_dir,
    *,
    four_c_exe=None,
    mpi_command=None,
    n_proc=None,
    output_name="xxx",
    restart_step=None,
    restart_from=None,
    log_to_console=False,
):
    """Run a 4C simulation and return the exit code of the run.

    This function looks into the environment variables for some parameters:
        "BEAMME_FOUR_C_EXE"
        "BEAMME_MPI_COMMAND"
        "BEAMME_MPI_NUM_PROC"
    If the corresponding keyword arguments are set, they overwrite the environment
    variable.

    Args
    ----
    input_file: str
        Path to the input file on the filesystem
    output_dir: str
        Directory where the simulation should be performed (will be created if
        it does not exist)
    four_c_exe: str
        Optionally explicitly specify path to the 4C executable
    mpi_command: str
        Command to launch MPI, defaults to "mpirun"
    n_proc: int
        Number of process used with MPI, defaults to 1
    output_name: str
        Base name of the output files
    restart_step: int
        Time step to restart from
    restart_from: str
        Path to initial simulation (relative to output_dir)
    log_to_console: bool
        If the 4C simulation output should be shown in the console.

    Return
    ----
    return_code: int
        Return code of 4C run
    """

    # Fist get all needed parameters
    if four_c_exe is None:
        four_c_exe = _get_env_variable("BEAMME_FOUR_C_EXE")
    if mpi_command is None:
        mpi_command = _get_env_variable("BEAMME_MPI_COMMAND", default="mpirun")
    if n_proc is None:
        n_proc = _get_env_variable("BEAMME_MPI_NUM_PROC", default="1")

    # Setup paths and actual command to run
    _os.makedirs(output_dir, exist_ok=True)
    log_file = _os.path.join(output_dir, output_name + ".log")
    error_file = _os.path.join(output_dir, output_name + ".err")
    command = mpi_command.split(" ") + [
        "-np",
        str(n_proc),
        four_c_exe,
        input_file,
        output_name,
    ]
    if restart_step is None and restart_from is None:
        pass
    elif restart_step is not None and restart_from is not None:
        command.extend([f"restart={restart_step}", f"restartfrom={restart_from}"])
    else:
        raise ValueError(
            "Provide either both or no argument of [restart_step, restart_from]"
        )

    # Actually run the command
    with open(log_file, "w") as stdout_file, open(error_file, "w") as stderr_file:
        process = _subprocess.Popen(
            command,  # nosec B603
            stdout=_subprocess.PIPE,
            stderr=_subprocess.PIPE,
            cwd=output_dir,
            text=True,
        )

        for stdout_line in process.stdout:
            if log_to_console:
                _sys.stdout.write(stdout_line)
            stdout_file.write(stdout_line)

        for stderr_line in process.stderr:
            if log_to_console:
                _sys.stderr.write(stderr_line)
            stderr_file.write(stderr_line)

        process.stdout.close()
        process.stderr.close()
        return_code = process.wait()
    return return_code


def clean_simulation_directory(sim_dir, *, ask_before_clean=False):
    """Clear the simulation directory. If it does not exist, it is created.
    Optionally the user can be asked before a deletion of files.

    Args
    ----
    sim_dir:
        Path to a directory
    ask_before_clean: bool
        Flag which indicates whether the user must confirm removal of files and directories
    """

    # Check if simulation directory exists.
    if _os.path.exists(sim_dir):
        if ask_before_clean:
            print(f'Path "{sim_dir}" already exists')
        while True:
            if ask_before_clean:
                answer = input("DELETE all contents? (y/n): ")
            else:
                answer = "y"
            if answer.lower() == "y":
                for filename in _os.listdir(sim_dir):
                    file_path = _os.path.join(sim_dir, filename)
                    try:
                        if _os.path.isfile(file_path) or _os.path.islink(file_path):
                            _os.unlink(file_path)
                        elif _os.path.isdir(file_path):
                            _shutil.rmtree(file_path)
                    except Exception as e:
                        raise ValueError(f"Failed to delete {file_path}. Reason: {e}")
                return
            elif answer.lower() == "n":
                raise ValueError("Directory is not deleted!")
    else:
        _Path(sim_dir).mkdir(parents=True, exist_ok=True)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file contains functionality to warp an existing mesh along a 1D
curve."""

from typing import Optional as _Optional
from typing import Tuple as _Tuple

import numpy as _np
import quaternion as _quaternion
from numpy.typing import NDArray as _NDArray

from beamme.core.boundary_condition import BoundaryCondition as _BoundaryCondition
from beamme.core.conf import bme as _bme
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.mesh import Mesh as _Mesh
from beamme.core.node import Node as _Node
from beamme.core.node import NodeCosserat as _NodeCosserat
from beamme.core.rotation import Rotation as _Rotation
from beamme.cosserat_curve.cosserat_curve import CosseratCurve as _CosseratCurve
from beamme.four_c.function_utility import (
    create_linear_interpolation_function as _create_linear_interpolation_function,
)
from beamme.geometric_search.find_close_points import (
    find_close_points as _find_close_points,
)


def get_arc_length_and_cross_section_coordinates(
    coordinates: _np.ndarray, origin: _np.ndarray, reference_rotation: _Rotation
) -> _Tuple[float, _np.ndarray]:
    """Return the arc length and the cross section coordinates for a coordinate
    system defined by the reference rotation and the origin.

    Args
    ----
    coordinates:
        Point coordinates in R3
    origin:
        Origin of the coordinate system
    reference_rotation:
        Rotation of the coordinate system. The first basis vector is the arc
        length direction.
    """

    transformed_coordinates = reference_rotation.inv() * (coordinates - origin)
    centerline_position = transformed_coordinates[0]
    cross_section_coordinates = [0.0, *transformed_coordinates[1:]]
    return centerline_position, cross_section_coordinates


def get_mesh_transformation(
    curve: _CosseratCurve,
    nodes: list[_Node],
    *,
    origin=[0.0, 0.0, 0.0],
    reference_rotation=_Rotation(),
    n_steps: int = 10,
    initial_configuration: bool = True,
    **kwargs,
) -> _Tuple[_np.ndarray, _NDArray[_quaternion.quaternion]]:
    """Generate a list of positions for each node that describe the
    transformation of the nodes from the given configuration to the Cosserat
    curve.

    Args
    ----
    curve:
        Curve to warp the mesh to
    nodes:
        Optional, if this is given only warp the given nodes. Per default all nodes in
        the mesh are warped.
    origin:
        Origin of the coordinate system
    reference_rotation:
        Rotation of the coordinate system. The first basis vector is the arc
        length direction.
    n_steps:
        Number of steps to get from the unwrapped configuration to the final configuration.
    initial_configuration:
        If the initial, unwrapped configuration (factor=0) should also be added to the
        results.
    kwargs:
        Keyword arguments passed to CosseratCurve.get_centerline_positions_and_rotations

    Return
    ----
    positions: list(_np.array(n_nodes x 3))
        A list for each time step containing the position of all nodes for that time step
    relative_rotations: list(list(Rotation))
        A list for each time step containing the relative rotations for all nodes at that
        time step
    """

    # Define the factors for which we will generate the positions and rotations
    factors = _np.linspace(0.0, 1.0, n_steps + 1)
    if initial_configuration:
        n_output_steps = n_steps + 1
    else:
        n_output_steps = n_steps
        factors = _np.delete(factors, 0)

    # Create output arrays
    n_nodes = len(nodes)
    positions = _np.zeros((n_output_steps, n_nodes, 3))
    relative_rotations = _np.zeros(
        (n_output_steps, n_nodes), dtype=_quaternion.quaternion
    )

    # Get all arc lengths and cross section positions
    arc_lengths = _np.zeros((n_nodes, 1))
    cross_section_coordinates = [None] * n_nodes
    for i_node, node in enumerate(nodes):
        (
            arc_lengths[i_node],
            cross_section_coordinates[i_node],
        ) = get_arc_length_and_cross_section_coordinates(
            node.coordinates, origin, reference_rotation
        )

    # Get unique arc length points
    has_partner, n_partner = _find_close_points(arc_lengths)
    arc_lengths_unique = [None] * n_partner
    has_partner_total = [-2] * len(arc_lengths)
    for i in range(len(arc_lengths)):
        partner_id = has_partner[i]
        if partner_id == -1:
            has_partner_total[i] = len(arc_lengths_unique)
            arc_lengths_unique.append(arc_lengths[i][0])
        else:
            if arc_lengths_unique[partner_id] is None:
                arc_lengths_unique[partner_id] = arc_lengths[i][0]
            has_partner_total[i] = partner_id

    n_total = len(arc_lengths_unique)
    arc_lengths_unique = _np.array(arc_lengths_unique)
    arc_lengths_sorted_index = _np.argsort(arc_lengths_unique)
    arc_lengths_sorted = arc_lengths_unique[arc_lengths_sorted_index]
    arc_lengths_sorted_index_inv = [-2 for i in range(n_total)]
    for i in range(n_total):
        arc_lengths_sorted_index_inv[arc_lengths_sorted_index[i]] = i
    point_to_unique = []
    for partner in has_partner_total:
        point_to_unique.append(arc_lengths_sorted_index_inv[partner])

    # Get all configurations for the unique points
    positions_for_all_steps = []
    quaternions_for_all_steps = []

    for factor in factors:
        sol_r, sol_q = curve.get_centerline_positions_and_rotations(
            arc_lengths_sorted, factor=factor, **kwargs
        )
        positions_for_all_steps.append(sol_r)
        quaternions_for_all_steps.append(sol_q)

    # Get data required for the rigid body motion
    curve_start_pos, curve_start_rot = curve.get_centerline_position_and_rotation(0.0)
    rigid_body_translation = curve_start_pos - origin
    rigid_body_rotation = curve_start_rot

    # Loop over nodes and map them to the new configuration
    for i_node, node in enumerate(nodes):
        if not isinstance(node, _Node):
            raise TypeError(
                "All nodes in the mesh have to be derived from the base Node object"
            )

        node_unique_id = point_to_unique[i_node]
        cross_section_position = cross_section_coordinates[i_node]

        # Check that the arc length coordinates match
        if (
            _np.abs(arc_lengths[i_node] - arc_lengths_sorted[node_unique_id])
            > _bme.eps_pos
        ):
            raise ValueError("Arc lengths do not match")

        # Create the functions that describe the deformation
        for i_step, factor in enumerate(factors):
            centerline_pos = positions_for_all_steps[i_step][node_unique_id]
            centerline_relative_pos = _quaternion.rotate_vectors(
                curve_start_rot.conjugate(), centerline_pos - curve_start_pos
            )
            centerline_rotation = quaternions_for_all_steps[i_step][node_unique_id]
            centerline_relative_rotation = (
                curve_start_rot.conjugate() * centerline_rotation
            )

            rigid_body_rotation_for_factor = _quaternion.slerp_evaluate(
                reference_rotation.get_numpy_quaternion(), rigid_body_rotation, factor
            )

            current_pos = (
                _quaternion.rotate_vectors(
                    rigid_body_rotation_for_factor,
                    (
                        centerline_relative_pos
                        + _quaternion.rotate_vectors(
                            centerline_relative_rotation, cross_section_position
                        )
                    ),
                )
                + origin
                + factor * rigid_body_translation
            )

            positions[i_step, i_node] = current_pos
            relative_rotations[i_step, i_node] = (
                rigid_body_rotation_for_factor
                * centerline_relative_rotation
                * reference_rotation.get_numpy_quaternion().conjugate()
            )

    return positions, relative_rotations


def create_transform_boundary_conditions(
    mesh: _Mesh,
    curve: _CosseratCurve,
    *,
    nodes: _Optional[list[_Node]] = None,
    t_end: float = 1.0,
    n_steps: int = 10,
    n_dof_per_node: int = 3,
    **kwargs,
) -> None:
    """Create the Dirichlet boundary conditions that enforce the warping. The
    warped object is assumed to align with the x-axis in the reference
    configuration.

    Args
    ----
    mesh:
        Mesh to be warped
    curve:
        Curve to warp the mesh to
    nodes:
        Optional, if this is given only warp the given nodes. Per default all nodes in
        the mesh are warped.
    n_steps:
        Number of steps to apply the warping condition
    t_end:
        End time for applying the warping boundary conditions
    n_dof_per_node:
        Number of DOF per node in 4C (is needed to correctly define the boundary conditions)
    kwargs:
        Keyword arguments passed to get_mesh_transformation
    """

    # If no nodes are given, use all nodes in the mesh
    if nodes is None:
        nodes = mesh.nodes

    time_values = _np.linspace(0.0, t_end, n_steps + 1)

    # Get positions and rotations for each step
    positions, _ = get_mesh_transformation(curve, nodes, n_steps=n_steps, **kwargs)

    # Loop over nodes and map them to the new configuration
    for i_node, node in enumerate(nodes):
        # Create the functions that describe the deformation
        reference_position = node.coordinates
        displacement_values = _np.array(
            [
                positions[i_step][i_node] - reference_position
                for i_step in range(n_steps + 1)
            ]
        )
        fun_pos = [
            _create_linear_interpolation_function(
                time_values, displacement_values[:, i_dir]
            )
            for i_dir in range(3)
        ]
        for fun in fun_pos:
            mesh.add(fun)
        n_additional_dof = n_dof_per_node - 3
        mesh.add(
            _BoundaryCondition(
                _GeometrySet(node),
                {
                    "NUMDOF": n_dof_per_node,
                    "ONOFF": [1] * 3 + [0] * n_additional_dof,
                    "VAL": [1.0] * 3 + [0.0] * n_additional_dof,
                    "FUNCT": fun_pos + [None] * n_additional_dof,
                    "TAG": "monitor_reaction",
                },
                bc_type=_bme.bc.dirichlet,
            )
        )


def warp_mesh_along_curve(
    mesh: _Mesh,
    curve: _CosseratCurve,
    *,
    origin=[0.0, 0.0, 0.0],
    reference_rotation=_Rotation(),
) -> None:
    """Warp an existing mesh along the given curve.

    The reference coordinates for the transformation are defined by the
    given origin and rotation, where the first basis vector of the triad
    defines the centerline axis.
    """

    pos, rot = get_mesh_transformation(
        curve,
        mesh.nodes,
        origin=origin,
        reference_rotation=reference_rotation,
        n_steps=1,
        initial_configuration=False,
    )

    # Loop over nodes and map them to the new configuration
    for i_node, node in enumerate(mesh.nodes):
        if not isinstance(node, _Node):
            raise TypeError(
                "All nodes in the mesh have to be derived from the base Node object"
            )

        new_pos = pos[0, i_node]
        node.coordinates = new_pos
        if isinstance(node, _NodeCosserat):
            node.rotation = _Rotation.from_quaternion(rot[0, i_node]) * node.rotation



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script contains functionality to create solid input files (or plain
cubit instances) with CubitPy which are then used in testing."""

from beamme.core.boundary_condition import BoundaryCondition
from beamme.core.conf import bme
from beamme.core.geometry_set import GeometrySet
from beamme.core.mesh import Mesh
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.four_c.input_file import InputFile
from beamme.four_c.model_importer import import_cubitpy_model, import_four_c_model
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line
from beamme.utils.environment import cubitpy_is_available

if cubitpy_is_available():
    from cubitpy import CubitPy, cupy
    from cubitpy.mesh_creation_functions import (
        create_brick,
        extrude_mesh_normal_to_surface,
    )


def create_tube_cubit_mesh(r, h, n_circumference, n_height):
    """Create a solid tube in cubit.

    Args
    ----
    r: float
        Radius of the cylinder.
    h: float
        Height of the cylinder.
    n_circumference: int
        Number of elements along the circumferential direction.
    n_height: int
        Number of elements along the axial direction.

    Return
    ----
    The created cubit object.
    """

    # Initialize cubit.
    cubit = CubitPy()

    # Create cylinder.
    cylinder = cubit.cylinder(h, r, r, r)

    # Set the mesh size.
    for curve in cylinder.curves():
        cubit.set_line_interval(curve, n_circumference)
    cubit.cmd("surface 1 size {}".format(h / n_height))

    # Set blocks and sets.
    cubit.add_element_type(cylinder.volumes()[0], cupy.element_type.hex8, name="tube")

    # Return the cubit object.
    return cubit, cylinder


def create_tube_cubit():
    """Load the solid tube and add input file parameters."""

    # Initialize cubit.
    cubit, cylinder = create_tube_cubit_mesh(0.25, 10.0, 6, 10)

    # Mesh the geometry.
    cylinder.volumes()[0].mesh()

    # Set boundary conditions.
    cubit.add_node_set(
        cylinder.surfaces()[1],
        name="fix",
        bc_section="DESIGN SURF DIRICH CONDITIONS",
        bc_description={
            "NUMDOF": 3,
            "ONOFF": [1, 1, 1],
            "VAL": [0, 0, 0],
            "FUNCT": [0, 0, 0],
        },
    )
    cubit.add_node_set(
        cylinder.surfaces()[2],
        name="dirichlet_controlled",
        bc_section="DESIGN SURF DIRICH CONDITIONS",
        bc_description={
            "NUMDOF": 3,
            "ONOFF": [1, 1, 1],
            "VAL": [3.0, 3.0, 0],
            "FUNCT": [1, 2, 0],
        },
    )

    # Set header.
    cubit.fourc_input.combine_sections(
        {
            "PROBLEM TYPE": {"PROBLEMTYPE": "Structure"},
            "IO": {
                "VERBOSITY": "Standard",
            },
            "IO/RUNTIME VTK OUTPUT": {
                "OUTPUT_DATA_FORMAT": "binary",
                "INTERVAL_STEPS": 1,
                "EVERY_ITERATION": False,
            },
            "IO/RUNTIME VTK OUTPUT/STRUCTURE": {
                "OUTPUT_STRUCTURE": True,
                "DISPLACEMENT": True,
            },
            "STRUCTURAL DYNAMIC": {
                "LINEAR_SOLVER": 1,
                "INT_STRATEGY": "Standard",
                "DYNAMICTYPE": "Statics",
                "RESTARTEVERY": 5,
                "PREDICT": "TangDis",
                "TIMESTEP": 0.05,
                "NUMSTEP": 20,
                "MAXTIME": 1.0,
                "TOLRES": 1.0e-5,
                "TOLDISP": 1.0e-11,
                "MAXITER": 20,
            },
            "SOLVER 1": {"NAME": "Structure_Solver", "SOLVER": "UMFPACK"},
            "MATERIALS": [
                {
                    "MAT": 1,
                    "MAT_Struct_StVenantKirchhoff": {
                        "YOUNG": 1.0e9,
                        "NUE": 0.0,
                        "DENS": 7.8e-6,
                    },
                }
            ],
            "FUNCT1": [
                {
                    "COMPONENT": 0,
                    "SYMBOLIC_FUNCTION_OF_SPACE_TIME": "cos(2*pi*t)",
                }
            ],
            "FUNCT2": [
                {
                    "COMPONENT": 0,
                    "SYMBOLIC_FUNCTION_OF_SPACE_TIME": "sin(2*pi*t)",
                }
            ],
        }
    )

    # Return the cubit object.
    return cubit


def create_tube(file_path):
    """Write the solid tube to a file."""

    # Export mesh.
    create_tube_cubit().dump(file_path)


def create_block_cubit():
    """Create a solid block in cubit and add a volume condition."""

    # Initialize cubit.
    cubit = CubitPy()

    # Create the block.
    cube = create_brick(cubit, 1, 1, 1, mesh_factor=9)

    # Set the material.
    cubit.fourc_input["MATERIALS"] = [
        {"MAT": 1, "MAT_Struct_StVenantKirchhoff": {"DENS": 1, "NUE": 0.3, "YOUNG": 2}}
    ]

    # Add the boundary condition.
    cubit.add_node_set(
        cube.volumes()[0],
        bc_type=cupy.bc_type.beam_to_solid_volume_meshtying,
        bc_description={"COUPLING_ID": 1},
    )

    # Add the boundary condition.
    cubit.add_node_set(
        cube.surfaces()[0],
        bc_type=cupy.bc_type.beam_to_solid_surface_meshtying,
        bc_description={"COUPLING_ID": 2},
    )

    # Set point coupling conditions.
    nodes = cubit.group()
    nodes.add([cube.vertices()[0], cube.vertices()[2]])
    cubit.add_node_set(
        nodes,
        bc_type=cupy.bc_type.point_coupling,
        bc_description={
            "NUMDOF": 3,
            "ONOFF": [1, 2, 3],
        },
    )

    # Return the cubit object.
    return cubit


def create_block(file_path):
    """Create the solid cube in cubit and write it to a file."""

    # Export mesh.
    create_block_cubit().dump(file_path)


def create_solid_shell_meshes(file_path_blocks, file_path_dome):
    """Create the meshes needed for the solid shell tests."""

    def create_brick_mesh(
        dimensions, n_elements, *, element_type=cupy.element_type.hex8sh
    ):
        """Create a mesh with a solid brick."""
        cubit = CubitPy()
        create_brick(
            cubit,
            *dimensions,
            mesh_interval=n_elements,
            element_type=element_type,
            mesh=True,
        )
        cubit.fourc_input["MATERIALS"] = [
            {
                "MAT": 1,
                "MAT_Struct_StVenantKirchhoff": {"DENS": 1, "NUE": 0.3, "YOUNG": 2},
            }
        ]
        _, mesh = import_cubitpy_model(cubit, convert_input_to_mesh=True)
        return mesh

    # Create the input file with the blocks representing plates in different planes
    input_file = InputFile()
    dimensions = [0.1, 2, 4]
    elements = [1, 2, 2]

    def rotate_list(original_list, n):
        """Rotate the list."""
        return original_list[-n:] + original_list[:-n]

    # Add the plates in all directions (permute the dimension and number of elements
    # in each direction)
    for i in range(3):
        brick = create_brick_mesh(rotate_list(dimensions, i), rotate_list(elements, i))
        brick.translate([i * 4, 0, 0])
        input_file.add(brick)

    # Add a last plate with standard solid elements, to make sure that the algorithm
    # skips those
    brick = create_brick_mesh(
        rotate_list(dimensions, 1),
        rotate_list(elements, 1),
        element_type=cupy.element_type.hex8,
    )
    brick.translate([3 * 4, 0, 0])
    input_file.add(brick)

    input_file.dump(
        file_path_blocks, add_header_information=False, validate_sections_only=True
    )

    # Create the dome input
    cubit = CubitPy()
    cubit.cmd("create sphere radius 1 zpositive")
    cubit.cmd("surface 2 size auto factor 6")
    cubit.cmd("mesh surface 2")
    dome_mesh = extrude_mesh_normal_to_surface(
        cubit, [cubit.surface(2)], 0.1, n_layer=1
    )
    cubit.add_element_type(dome_mesh, cupy.element_type.hex8sh)
    cubit.fourc_input["MATERIALS"] = [
        {
            "MAT": 1,
            "MAT_Struct_StVenantKirchhoff": {"DENS": 1, "NUE": 0.3, "YOUNG": 2},
        }
    ]
    cubit.dump(file_path_dome)


def create_beam_to_solid_conditions_model(
    get_default_test_beam_material,
    get_corresponding_reference_file_path,
    full_import: bool,
):
    """Create the input file for the beam-to-solid input conditions tests."""

    # Create input file
    input_file, mesh = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            reference_file_base_name="test_other_create_cubit_input_files_block"
        ),
        convert_input_to_mesh=full_import,
    )

    # Add beams to the model
    mesh_beams = Mesh()
    material = get_default_test_beam_material(material_type="reissner")
    create_beam_mesh_line(
        mesh_beams, Beam3rHerm2Line3, material, [0, 0, 0], [0, 0, 1], n_el=3
    )
    create_beam_mesh_line(
        mesh_beams, Beam3rHerm2Line3, material, [0, 0.5, 0], [0, 0.5, 1], n_el=3
    )

    # Set beam-to-solid coupling conditions.
    line_set = GeometrySet(mesh_beams.elements)
    mesh_beams.add(
        BoundaryCondition(
            line_set,
            bc_type=bme.bc.beam_to_solid_volume_meshtying,
            data={"COUPLING_ID": 1},
        )
    )
    mesh_beams.add(
        BoundaryCondition(
            line_set,
            bc_type=bme.bc.beam_to_solid_surface_meshtying,
            data={"COUPLING_ID": 2},
        )
    )
    mesh.add(mesh_beams)

    return input_file, mesh


def create_single_solid_element_brick(input_file_path, get_default_test_solid_material):
    """Create an input file with a single solid element brick in CubitPy for
    testing purposes."""

    # Create the brick with a single solid element
    cubit = CubitPy()
    create_brick(
        cubit, 1, 2, 3, mesh_interval=[1, 1, 1], element_type=cupy.element_type.hex8
    )
    material = get_default_test_solid_material(material_type="st_venant_kirchhoff")
    material.i_global = 0
    cubit.fourc_input["MATERIALS"] = [material.dump_to_list()]
    cubit.dump(input_file_path)
